@@ -98,7 +98,7 @@ static const char *preen_msg[] = {
 	"",			/* 20 */
 };
 
-static const struct e2fsck_problem problem_table[] = {
+static struct e2fsck_problem problem_table[] = {
 
 	/* Pre-Pass 1 errors */
 
@@ -332,6 +332,20 @@ static const struct e2fsck_problem problem_table[] = {
 	  N_("Resize @i not valid.  "),
 	  PROMPT_RECREATE, 0 },
 
+	/* Last mount time is in the future */
+	{ PR_0_FUTURE_SB_LAST_MOUNT,
+	  N_("@S last mount time is in the future.  "),
+	  PROMPT_FIX, PR_PREEN_OK },
+
+	/* Last write time is in the future */
+	{ PR_0_FUTURE_SB_LAST_WRITE,
+	  N_("@S last write time is in the future.  "),
+	  PROMPT_FIX, PR_PREEN_OK },
+
+	{ PR_0_EXTERNAL_JOURNAL_HINT,
+	  N_("@S hint for external superblock @s %X.  "),
+	     PROMPT_FIX, PR_PREEN_OK },
+
 	/* Pass 1 errors */
 	
 	/* Pass 1: Checking inodes, blocks, and sizes */
@@ -806,7 +820,7 @@ static const struct e2fsck_problem problem_table[] = {
 
 	/* Pass 1C: Scan directories for inodes with multiply-claimed blocks. */
 	{ PR_1C_PASS_HEADER,
-	  N_("Pass 1C: Scanning directories for @is with @m @bs.\n"),
+	  N_("Pass 1C: Scanning directories for @is with @m @bs\n"),
 	  PROMPT_NONE, 0 },
 
 		  
@@ -1487,7 +1501,7 @@ static struct latch_descr pr_latch_info[] = {
 	{ -1, 0, 0 },
 };
 
-static const struct e2fsck_problem *find_problem(problem_t code)
+static struct e2fsck_problem *find_problem(problem_t code)
 {
 	int 	i;
 
@@ -1554,10 +1568,24 @@ void clear_problem_context(struct problem_context *ctx)
 	ctx->group = -1;
 }
 
+static void reconfigure_bool(e2fsck_t ctx, struct e2fsck_problem *ptr, 
+			     const char *key, int mask, const char *name)
+{
+	int	bool;
+
+	bool = (ptr->flags & mask);
+	profile_get_boolean(ctx->profile, "problems", key, name, bool, &bool);
+	if (bool)
+		ptr->flags |= mask;
+	else
+		ptr->flags &= ~mask;
+}
+
+
 int fix_problem(e2fsck_t ctx, problem_t code, struct problem_context *pctx)
 {
 	ext2_filsys fs = ctx->fs;
-	const struct e2fsck_problem *ptr;
+	struct e2fsck_problem *ptr;
 	struct latch_descr *ldesc = 0;
 	const char *message;
 	int 		def_yn, answer, ans;
@@ -1569,6 +1597,27 @@ int fix_problem(e2fsck_t ctx, problem_t code, struct problem_context *pctx)
 		printf(_("Unhandled error code (0x%x)!\n"), code);
 		return 0;
 	}
+	if (!(ptr->flags & PR_CONFIG)) {
+		char	key[9], *new_desc;
+
+		sprintf(key, "0x%06x", code);
+
+		profile_get_string(ctx->profile, "problems", key,
+				   "description", 0, &new_desc);
+		if (new_desc)
+			ptr->e2p_description = new_desc;
+
+		reconfigure_bool(ctx, ptr, key, PR_PREEN_OK, "preen_ok");
+		reconfigure_bool(ctx, ptr, key, PR_NO_OK, "no_ok");
+		reconfigure_bool(ctx, ptr, key, PR_NO_DEFAULT, "no_default");
+		reconfigure_bool(ctx, ptr, key, PR_MSG_ONLY, "print_message_only");
+		reconfigure_bool(ctx, ptr, key, PR_PREEN_NOMSG, "preen_nomessage");
+		reconfigure_bool(ctx, ptr, key, PR_NOCOLLATE, "no_collate");
+		reconfigure_bool(ctx, ptr, key, PR_NO_NOMSG, "no_nomsg");
+		reconfigure_bool(ctx, ptr, key, PR_PREEN_NOHDR, "preen_noheader");
+
+		ptr->flags |= PR_CONFIG;
+	}
 	def_yn = 1;
 	if ((ptr->flags & PR_NO_DEFAULT) ||
 	    ((ptr->flags & PR_PREEN_NO) && (ctx->options & E2F_OPT_PREEN)) ||
