@@ -39,6 +39,12 @@ extern int optind;
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
 #endif
+#ifdef HAVE_SYS_TYPES_H
+#include <sys/types.h>
+#endif
+#ifdef HAVE_DIRENT_H
+#include <dirent.h>
+#endif
 
 #include "et/com_err.h"
 #include "e2fsck.h"
@@ -67,7 +73,7 @@ static void usage(e2fsck_t ctx)
 	fprintf(stderr,
 		_("Usage: %s [-panyrcdfvstDFSV] [-b superblock] [-B blocksize]\n"
 		"\t\t[-I inode_buffer_blocks] [-P process_inode_size]\n"
-		"\t\t[-l|-L bad_blocks_file] [-C fd] [-j ext-journal]\n"
+		"\t\t[-l|-L bad_blocks_file] [-C fd] [-j external_journal]\n"
 		"\t\t[-E extended-options] device\n"),
 		ctx->program_name);
 
@@ -81,7 +87,7 @@ static void usage(e2fsck_t ctx)
 		" -v                   Be verbose\n"
 		" -b superblock        Use alternative superblock\n"
 		" -B blocksize         Force blocksize when looking for superblock\n"
-		" -j external-journal  Set location of the external journal\n"
+		" -j external_journal  Set location of the external journal\n"
 		" -l bad_blocks_file   Add to badblocks list\n"
 		" -L bad_blocks_file   Set badblocks list\n"
 		));
@@ -92,7 +98,8 @@ static void usage(e2fsck_t ctx)
 static void show_stats(e2fsck_t	ctx)
 {
 	ext2_filsys fs = ctx->fs;
-	int inodes, inodes_used, blocks, blocks_used;
+	int inodes, inodes_used;
+	blk_t blocks, blocks_used;
 	int dir_links;
 	int num_files, num_links;
 	int frag_percent;
@@ -111,7 +118,7 @@ static void show_stats(e2fsck_t	ctx)
 	frag_percent = (frag_percent + 5) / 10;
 	
 	if (!verbose) {
-		printf(_("%s: %d/%d files (%0d.%d%% non-contiguous), %d/%d blocks\n"),
+		printf(_("%s: %d/%d files (%0d.%d%% non-contiguous), %u/%u blocks\n"),
 		       ctx->device_name, inodes_used, inodes,
 		       frag_percent / 10, frag_percent % 10,
 		       blocks_used, blocks);
@@ -125,7 +132,7 @@ static void show_stats(e2fsck_t	ctx)
 		ctx->fs_fragmented, frag_percent / 10, frag_percent % 10);
 	printf (_("         # of inodes with ind/dind/tind blocks: %d/%d/%d\n"),
 		ctx->fs_ind_count, ctx->fs_dind_count, ctx->fs_tind_count);
-	printf (P_("%8d block used (%d%%)\n", "%8d blocks used (%d%%)\n",
+	printf (P_("%8u block used (%d%%)\n", "%8u blocks used (%d%%)\n",
 		   blocks_used),
 		blocks_used, (int) ((long long) 100 * blocks_used / blocks));
 	printf (P_("%8d bad block\n", "%8d bad blocks\n",
@@ -173,15 +180,18 @@ static void check_mount(e2fsck_t ctx)
 	}
 
 	/*
-	 * If the filesystem isn't mounted, or it's the root filesystem
-	 * and it's mounted read-only, then everything's fine.
+	 * If the filesystem isn't mounted, or it's the root
+	 * filesystem and it's mounted read-only, and we're not doing
+	 * a read/write check, then everything's fine.
 	 */
 	if ((!(ctx->mount_flags & EXT2_MF_MOUNTED)) ||
 	    ((ctx->mount_flags & EXT2_MF_ISROOT) &&
-	     (ctx->mount_flags & EXT2_MF_READONLY)))
+	     (ctx->mount_flags & EXT2_MF_READONLY) &&
+	     !(ctx->options & E2F_OPT_WRITECHECK)))
 		return;
 
-	if (ctx->options & E2F_OPT_READONLY) {
+	if ((ctx->options & E2F_OPT_READONLY) &&
+	    !(ctx->options & E2F_OPT_WRITECHECK)) {
 		printf(_("Warning!  %s is mounted.\n"), ctx->filesystem_name);
 		return;
 	}
@@ -203,8 +213,10 @@ static void check_mount(e2fsck_t ctx)
 static int is_on_batt(void)
 {
 	FILE	*f;
-	char	tmp[80], tmp2[80];
+	DIR	*d;
+	char	tmp[80], tmp2[80], fname[80];
 	unsigned int	acflag;
+	struct dirent*	de;
 
 	f = fopen("/proc/apm", "r");
 	if (f) {
@@ -213,13 +225,25 @@ static int is_on_batt(void)
 		fclose(f);
 		return (acflag != 1);
 	}
-	f = fopen("/proc/acpi/ac_adapter/AC/state", "r");
-	if (f) {
-		if (fscanf(f, "%s %s", tmp2, tmp) != 2)
-			tmp[0] = 0;
-		fclose(f);
-		if (strncmp(tmp, "off-line", 8) == 0)
-			return 1;
+	d = opendir("/proc/acpi/ac_adapter");
+	if (d) {
+		while ((de=readdir(d)) != NULL) {
+			if (!strncmp(".", de->d_name, 1))
+				continue;
+			snprintf(fname, 80, "/proc/acpi/ac_adapter/%s/state", 
+				 de->d_name);
+			f = fopen(fname, "r");
+			if (!f)
+				continue;
+			if (fscanf(f, "%s %s", tmp2, tmp) != 2)
+				tmp[0] = 0;
+			fclose(f);
+			if (strncmp(tmp, "off-line", 8) == 0) {
+				closedir(d);
+				return 1;
+			}
+		}
+		closedir(d);
 	}
 	return 0;
 }
@@ -236,13 +260,20 @@ static void check_if_skip(e2fsck_t ctx)
 	unsigned int reason_arg = 0;
 	long next_check;
 	int batt = is_on_batt();
-	time_t now = time(0);
-	
+	int defer_check_on_battery;
+
+	profile_get_boolean(ctx->profile, "options",
+			    "defer_check_on_battery", 0, 1, 
+			    &defer_check_on_battery);
+	if (!defer_check_on_battery)
+		batt = 0;
+
 	if ((ctx->options & E2F_OPT_FORCE) || bad_blocks_file ||
 	    cflag || swapfs)
 		return;
 	
-	if (fs->super->s_state & EXT2_ERROR_FS)
+	if ((fs->super->s_state & EXT2_ERROR_FS) ||
+	    !ext2fs_test_valid(fs))
 		reason = _(" contains a file system with errors");
 	else if ((fs->super->s_state & EXT2_VALID_FS) == 0)
 		reason = _(" was not cleanly unmounted");
@@ -255,11 +286,11 @@ static void check_if_skip(e2fsck_t ctx)
 			     (unsigned) fs->super->s_max_mnt_count*2))
 			reason = 0;
 	} else if (fs->super->s_checkinterval &&
-		   ((now - fs->super->s_lastcheck) >= 
+		   ((ctx->now - fs->super->s_lastcheck) >= 
 		    fs->super->s_checkinterval)) {
 		reason = _(" has gone %u days without being checked");
-		reason_arg = (now - fs->super->s_lastcheck)/(3600*24);
-		if (batt && ((now - fs->super->s_lastcheck) < 
+		reason_arg = (ctx->now - fs->super->s_lastcheck)/(3600*24);
+		if (batt && ((ctx->now - fs->super->s_lastcheck) < 
 			     fs->super->s_checkinterval*2))
 			reason = 0;
 	}
@@ -269,7 +300,7 @@ static void check_if_skip(e2fsck_t ctx)
 		fputs(_(", check forced.\n"), stdout);
 		return;
 	}
-	printf(_("%s: clean, %d/%d files, %d/%d blocks"), ctx->device_name,
+	printf(_("%s: clean, %d/%d files, %u/%u blocks"), ctx->device_name,
 	       fs->super->s_inodes_count - fs->super->s_free_inodes_count,
 	       fs->super->s_inodes_count,
 	       fs->super->s_blocks_count - fs->super->s_free_blocks_count,
@@ -280,12 +311,17 @@ static void check_if_skip(e2fsck_t ctx)
 		if (next_check <= 0) 
 			next_check = 1;
 	}
-	if ((now - fs->super->s_lastcheck) >= fs->super->s_checkinterval)
+	if (fs->super->s_checkinterval &&
+	    ((ctx->now - fs->super->s_lastcheck) >= fs->super->s_checkinterval))
 		next_check = 1;
 	if (next_check <= 5) {
-		if (next_check == 1)
-			fputs(_(" (check after next mount)"), stdout);
-		else
+		if (next_check == 1) {
+			if (batt) 
+				fputs(_(" (check deferred; on battery)"),
+				      stdout);
+			else
+				fputs(_(" (check after next mount)"), stdout);
+		} else
 			printf(_(" (check in %ld mounts)"), next_check);
 	}
 	fputc('\n', stdout);
@@ -340,6 +376,7 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 	unsigned int	tick;
 	struct timeval	tv;
 	int dpywidth;
+	int fixed_percent;
 
 	if (ctx->flags & E2F_FLAG_PROG_SUPPRESS)
 		return 0;
@@ -349,9 +386,10 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 	 * percentage hasn't changed, then we skip out right
 	 * away. 
 	 */
-	if (ctx->progress_last_percent == (int) 10 * percent)
+	fixed_percent = (int) ((10 * percent) + 0.5);
+	if (ctx->progress_last_percent == fixed_percent)
 		return 0;
-	ctx->progress_last_percent = (int) 10 * percent;
+	ctx->progress_last_percent = fixed_percent;
 
 	/*
 	 * If we've already updated the spinner once within
@@ -361,7 +399,7 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 	gettimeofday(&tv, NULL);
 	tick = (tv.tv_sec << 3) + (tv.tv_usec / (1000000 / 8));
 	if ((tick == ctx->progress_last_time) &&
-	    (percent != 0.0) && (percent != 100.0))
+	    (fixed_percent != 0) && (fixed_percent != 1000))
 		return 0;
 	ctx->progress_last_time = tick;
 
@@ -381,16 +419,18 @@ int e2fsck_simple_progress(e2fsck_t ctx, const char *label, float percent,
 	printf("%s%s: |%s%s", ctx->start_meta, label,
 	       bar + (sizeof(bar) - (i+1)),
 	       spaces + (sizeof(spaces) - (dpywidth - i + 1)));
-	if (percent == 100.0)
+	if (fixed_percent == 1000)
 		fputc('|', stdout);
 	else
 		fputc(spinner[ctx->progress_pos & 3], stdout);
+	printf(" %4.1f%%  ", percent);
 	if (dpynum)
-		printf(" %4.1f%%  %u\r%s", percent, dpynum, ctx->stop_meta);
+		printf("%u\r", dpynum);
 	else
-		printf(" %4.1f%%   \r%s", percent, ctx->stop_meta);
+		fputs(" \r", stdout);
+	fputs(ctx->stop_meta, stdout);
 	
-	if (percent == 100.0)
+	if (fixed_percent == 1000)
 		e2fsck_clear_progbar(ctx);
 	fflush(stdout);
 
@@ -504,19 +544,31 @@ static void parse_extended_opts(e2fsck_t ctx, const char *opts)
 				continue;
 			}
 			ctx->ext_attr_ver = ea_ver;
-		} else
+		} else {
+			fprintf(stderr, _("Unknown extended option: %s\n"),
+				token);
 			extended_usage++;
+		}
 	}
 	if (extended_usage) {
-		fprintf(stderr, _("Extended options are separated by commas, "
-			"and may take an argument which\n"
-			"is set off by an equals ('=') sign.  "
-			"Valid raid options are:\n"
-			"\tea_ver=<ea_version (1 or 2)\n\n"));
+		fputs(("\nExtended options are separated by commas, "
+		       "and may take an argument which\n"
+		       "is set off by an equals ('=') sign.  "
+			"Valid extended options are:\n"
+		       "\tea_ver=<ea_version (1 or 2)>\n\n"), stderr);
 		exit(1);
 	}
 }	
 
+static void syntax_err_report(const char *filename, long err, int line_num)
+{
+	fprintf(stderr, 
+		_("Syntax error in e2fsck config file (%s, line #%d)\n\t%s\n"),
+		filename, line_num, error_message(err));
+	exit(FSCK_ERROR);
+}
+
+static const char *config_fn[] = { "/etc/e2fsck.conf", 0 };
 
 static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 {
@@ -531,6 +583,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	struct sigaction	sa;
 #endif
 	char		*extended_opts = 0;
+	char		*cp;
 
 	retval = e2fsck_allocate_context(&ctx);
 	if (retval)
@@ -549,6 +602,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	memset(bar, '=', sizeof(bar)-1);
 	memset(spaces, ' ', sizeof(spaces)-1);
 	initialize_ext2_error_table();
+	initialize_prof_error_table();
 	blkid_get_cache(&ctx->blkid, NULL);
 	
 	if (argc && *argv)
@@ -585,7 +639,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 			if (ctx->options & (E2F_OPT_YES|E2F_OPT_NO)) {
 			conflict_opt:
 				fatal_error(ctx, 
-	_("Only one the options -p/-a, -n or -y may be specified."));
+	_("Only one of the options -p/-a, -n or -y may be specified."));
 			}
 			ctx->options |= E2F_OPT_PREEN;
 			break;
@@ -689,6 +743,9 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	if ((ctx->options & E2F_OPT_NO) && !bad_blocks_file &&
 	    !cflag && !swapfs && !(ctx->options & E2F_OPT_COMPRESS_DIRS))
 		ctx->options |= E2F_OPT_READONLY;
+	ctx->io_options = strchr(argv[optind], '?');
+	if (ctx->io_options) 
+		*ctx->io_options++ = 0;
 	ctx->filesystem_name = blkid_get_devname(ctx->blkid, argv[optind], 0);
 	if (!ctx->filesystem_name) {
 		com_err(ctx->program_name, 0, _("Unable to resolve '%s'"), 
@@ -697,7 +754,12 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	}
 	if (extended_opts)
 		parse_extended_opts(ctx, extended_opts);
-	
+
+	if ((cp = getenv("E2FSCK_CONFIG")) != NULL)
+		config_fn[0] = cp;
+	profile_set_syntax_err_cb(syntax_err_report);
+	profile_init(config_fn, &ctx->profile);
+
 	if (flush) {
 		fd = open(ctx->filesystem_name, O_RDONLY, 0);
 		if (fd < 0) {
@@ -849,24 +911,27 @@ restart:
 	flags = 0;
 	if ((ctx->options & E2F_OPT_READONLY) == 0)
 		flags |= EXT2_FLAG_RW;
+	if ((ctx->mount_flags & EXT2_MF_MOUNTED) == 0)
+		flags |= EXT2_FLAG_EXCLUSIVE;
 
 	if (ctx->superblock && ctx->blocksize) {
-		retval = ext2fs_open(ctx->filesystem_name, flags,
-				     ctx->superblock, ctx->blocksize,
-				     io_ptr, &fs);
+		retval = ext2fs_open2(ctx->filesystem_name, ctx->io_options, 
+				      flags, ctx->superblock, ctx->blocksize,
+				      io_ptr, &fs);
 	} else if (ctx->superblock) {
 		int blocksize;
 		for (blocksize = EXT2_MIN_BLOCK_SIZE;
 		     blocksize <= EXT2_MAX_BLOCK_SIZE; blocksize *= 2) {
-			retval = ext2fs_open(ctx->filesystem_name, flags,
-					     ctx->superblock, blocksize,
-					     io_ptr, &fs);
+			retval = ext2fs_open2(ctx->filesystem_name, 
+					      ctx->io_options, flags,
+					      ctx->superblock, blocksize,
+					      io_ptr, &fs);
 			if (!retval)
 				break;
 		}
 	} else 
-		retval = ext2fs_open(ctx->filesystem_name, flags, 
-				     0, 0, io_ptr, &fs);
+		retval = ext2fs_open2(ctx->filesystem_name, ctx->io_options, 
+				      flags, 0, 0, io_ptr, &fs);
 	if (!ctx->superblock && !(ctx->options & E2F_OPT_PREEN) &&
 	    !(ctx->flags & E2F_FLAG_SB_SPECIFIED) &&
 	    ((retval == EXT2_ET_BAD_MAGIC) ||
@@ -899,6 +964,9 @@ restart:
 			       "r/o" : "r/w");
 		else if (retval == ENXIO)
 			printf(_("Possibly non-existent or swap device?\n"));
+		else if (retval == EBUSY)
+			printf(_("Filesystem mounted or opened exclusively "
+				 "by another program?\n"));
 #ifdef EROFS
 		else if (retval == EROFS)
 			printf(_("Disk write-protected; use the -n option "
@@ -911,6 +979,7 @@ restart:
 	}
 	ctx->fs = fs;
 	fs->priv_data = ctx;
+	fs->now = ctx->now;
 	sb = fs->super;
 	if (sb->s_rev_level > E2FSCK_CURRENT_REV) {
 		com_err(ctx->program_name, EXT2_ET_REV_TOO_HIGH,
@@ -1031,6 +1100,7 @@ restart:
 
 	if (ctx->superblock)
 		set_latch_flags(PR_LATCH_RELOC, PRL_LATCHED, 0);
+	ext2fs_mark_valid(fs);
 	check_super_block(ctx);
 	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
 		fatal_error(ctx, 0);
@@ -1106,15 +1176,26 @@ restart:
 			exit_value |= FSCK_REBOOT;
 		}
 	}
-	if (!ext2fs_test_valid(fs)) {
+	if (!ext2fs_test_valid(fs) ||
+	    ((exit_value & FSCK_CANCELED) && 
+	     (sb->s_state & EXT2_ERROR_FS))) {
 		printf(_("\n%s: ********** WARNING: Filesystem still has "
 			 "errors **********\n\n"), ctx->device_name);
 		exit_value |= FSCK_UNCORRECTED;
 		exit_value &= ~FSCK_NONDESTRUCT;
 	}
-	if (exit_value & FSCK_CANCELED)
+	if (exit_value & FSCK_CANCELED) {
+		int	allow_cancellation;
+
+		profile_get_boolean(ctx->profile, "options",
+				    "allow_cancellation", 0, 0, 
+				    &allow_cancellation);
 		exit_value &= ~FSCK_NONDESTRUCT;
-	else {
+		if (allow_cancellation && ext2fs_test_valid(fs) &&
+		    (sb->s_state & EXT2_VALID_FS) && 
+		    !(sb->s_state & EXT2_ERROR_FS))
+			exit_value = 0;
+	} else {
 		show_stats(ctx);
 		if (!(ctx->options & E2F_OPT_READONLY)) {
 			if (ext2fs_test_valid(fs)) {
@@ -1124,7 +1205,7 @@ restart:
 			} else
 				sb->s_state &= ~EXT2_VALID_FS;
 			sb->s_mnt_count = 0;
-			sb->s_lastcheck = time(NULL);
+			sb->s_lastcheck = ctx->now;
 			ext2fs_mark_super_dirty(fs);
 		}
 	}
@@ -1135,12 +1216,11 @@ restart:
 	ctx->fs = NULL;
 	free(ctx->filesystem_name);
 	free(ctx->journal_name);
-	e2fsck_free_context(ctx);
-	
+
 #ifdef RESOURCE_TRACK
 	if (ctx->options & E2F_OPT_TIME)
 		print_resource_track(NULL, &ctx->global_rtrack);
 #endif
-
+	e2fsck_free_context(ctx);
 	return exit_value;
 }
